import logging
from copy import copy
from dataclasses import dataclass
from typing import Optional

import numpy as np
import matplotlib.pyplot as pt

@dataclass
class Settings:
    marker : int = 0
    awg_offs0 : int = 0
    awg_offs1 : int = 0
    awg_gain0 : int = 0
    awg_gain1 : int = 0
    phase: Optional[float] = None
    phase_shift : float = 0

def _phase2float(arg0, arg1, arg2):
    phase = ((arg2/6250 + arg1)/400 + arg0)/400
#    print(f'Phase {phase:6.3f}')
    return phase

def float2int16array(value):
    return np.array(value*2**15, dtype=np.int32)

def _i16(value):
    return np.int16(value)

class Renderer:

    def __init__(self, name):
        self.name = name
        self.max_render_time = 2_000_000
        self.wavedict = {}
        self.acq_weights = {}
        self.acq_bins = {}
        self.path_out_enabled = [set(), set()]
        self.reset()

    def reset(self):
        self.settings = Settings()
        self.next_settings = Settings()
        self.time = 0
        self.wave_start = 0
        self.waves_end = (0,0)
        self.waves = (None, None)
        self.nco_frequency = 0.0
        self.nco_enabled = False
        self.nco_phase_offset = 0.0
        self.out0 = []
        self.out1 = []
        self.acq_count = {
                i:[0]*num_bins for i,num_bins in self.acq_bins.items()
                }
        self.acq_data = {
                i:[float('nan')]*num_bins for i,num_bins in self.acq_bins.items()
                }
        self.acq_times = {i:[] for i in self.acq_bins}
        self.acq_buffer = AcqBuffer()
        self.errors = set()

    def path_enable(self, path, out, enable):
        if enable:
            self.path_out_enabled[path].add(out)
        else:
            self.path_out_enabled[path].discard(out)

    def set_nco(self, frequency, enabled):
        self.nco_frequency = frequency
        self.nco_enabled = enabled

    def set_waveforms(self, wavedict):
        self.wavedict = wavedict

    def set_weights(self, weightsdict):
        self.acq_weights = weightsdict

    def set_acquisition_bins(self, acq_bins):
        self.acq_bins = acq_bins

    def set_mrk(self, value):
        self.next_settings.marker = value

    def reset_ph(self):
        self.next_settings.phase = 0.0

    def set_ph(self, arg0, arg1, arg2):
        self.next_settings.phase = _phase2float(arg0, arg1, arg2)

    def set_ph_delta(self, arg0, arg1, arg2):
        self.next_settings.phase_shift = _phase2float(arg0, arg1, arg2)

    def set_awg_gain(self, gain0, gain1):
        self.next_settings.awg_gain0 = gain0
        self.next_settings.awg_gain1 = gain1

    def set_awg_offs(self, offset0, offset1):
        self.next_settings.awg_offs0 = offset0
        self.next_settings.awg_offs1 = offset1

    def upd_param(self, wait_after):
        self._update_settings()
        self._render(wait_after)

    def play(self, wave0, wave1, wait_after):
        self._update_settings()
        if wave0 not in self.wavedict:
            self._error('AWG WAVE PLAYBACK INDEX INVALID PATH 0')
        elif wave1 not in self.wavedict:
            self._error('AWG WAVE PLAYBACK INDEX INVALID PATH 1')
        else:
            self.wave_start = self.time
            self.waves = (float2int16array(self.wavedict[wave0]),
                          float2int16array(self.wavedict[wave1]))
            self.waves_end = (self.time + len(self.waves[0]),
                              self.time + len(self.waves[1]))
        self._render(wait_after)

    def acquire(self, bins, bin_index, wait_after):
        self._update_settings()
        self._add_acquisition(bins, bin_index)
        self._render(wait_after)

    def acquire_weighed(self, bins, bin_index, weight0, weight1, wait_after):
        self._update_settings()
        if weight0 not in self.acq_weights:
            self._error('ACQ WEIGHT PLAYBACK INDEX INVALID PATH 0')
        elif weight1 not in self.acq_weights:
            self._error('ACQ WEIGHT PLAYBACK INDEX INVALID PATH 1')
        else:
            self._add_acquisition(bins, bin_index)
        self._render(wait_after)

    def wait(self, time):
        self._render(time)

    def wait_sync(self, wait_after):
        self._render(wait_after)

    def _error(self, msg):
        logging.error(f'{self.name}: {msg}')
        self.errors.add(msg)

    def _update_settings(self):
        new = self.next_settings
        if new.phase is not None:
            phase = (new.phase - self.time * self.nco_frequency * 1e-9) % 1
            self.nco_phase_offset = phase
        self.nco_phase_offset += new.phase_shift
        self.settings = self.next_settings
        self.settings.phase = None
        self.next_settings = copy(self.settings)
        self.next_settings.phase = None
        self.next_settings.phase_shift = 0.0

    def _render(self, time):
        if time & 0x0003:
            logging.error(f'{self.name}: wait time not aligned on '
                          f'4 ns boundary: {time} ns (offset={time&0x03} ns)')
            self._error('TIME NOT ALIGNED')
        # 16 bits, 4 ns resolution
        time &= 0xFFFC
        t_start = self.time
        t_end = t_start + time
        self.time = t_end

        # stop rendering when there is too much data
        if t_start > self.max_render_time:
            return
        if t_end > self.max_render_time:
            t_end = self.max_render_time

        t_render = t_end - t_start

        s = self.settings

        path0 = np.full(t_render, s.awg_offs0, dtype=np.int16)
        path1 = np.full(t_render, s.awg_offs1, dtype=np.int16)

        if self.waves_end[0] > t_start:
            end = min(self.waves_end[0], t_end)
            data = self.waves[0][t_start-self.wave_start:end-self.wave_start]
            path0[0:len(data)] += (int(_i16(s.awg_gain0)) * data // 2**15)
        if self.waves_end[1] > t_start:
            end = min(self.waves_end[1], t_end)
            data = self.waves[1][t_start-self.wave_start:end-self.wave_start]
            path1[0:len(data)] += (int(_i16(s.awg_gain1)) * data // 2**15)

        if self.nco_enabled:
            t = np.arange(t_start, t_end)
            phase = (2*np.pi*self.nco_phase_offset + 2*np.pi*self.nco_frequency * 1e-9 * t)
            lo = np.cos(phase) + 1j*np.sin(phase)
            data = path0 + 1j*path1
            data = data * lo
            data0 = data.real.astype(np.int16)
            data1 = data.imag.astype(np.int16)
        else:
            data0 = path0
            data1 = path1

        if len(self.path_out_enabled[0]):
            self.out0.append(data0)
        if len(self.path_out_enabled[1]):
            self.out1.append(data1)

    def _add_acquisition(self, bins, bin_index):
        t = self.time
        if bins not in self.acq_bins:
            self._error('ACQ INDEX INVALID')
            return
        self.acq_times[bins].append((t, bin_index))
        if bin_index >= self.acq_bins[bins]:
            self._error('ACQ BIN INDEX INVALID')
        elif not self.acq_buffer.add(t):
            self._error('ACQ BINNING FIFO ERROR')
        else:
            if self.acq_count[bins][bin_index] == 0:
                self.acq_data[bins][bin_index] = t
            else:
                self.acq_data[bins][bin_index] += t
            self.acq_count[bins][bin_index] += 1

    def plot(self):
        if self.time > self.max_render_time:
            max_ms = self.max_render_time / 1e6
            print(f'{self.name}: Rendering truncated at {max_ms:3.1f} ms. Total time: {self.time/1e6:4.1f} ms')
        if len(self.path_out_enabled[0]):
            out0 = np.concatenate(self.out0)
            pt.plot(out0, label=f'{self.name}.{self.path_out_enabled[0]}')
        if len(self.path_out_enabled[1]):
            out1 = np.concatenate(self.out1)
            pt.plot(out1, label=f'{self.name}.{self.path_out_enabled[1]}')

    def get_acquisition_data(self):
        return (self.acq_count, self.acq_data)

    def get_acquisition_list(self):
        return self.acq_times

class AcqBuffer:
    def __init__(self):
        self.buffer = []
        self.write_ready = 0

    def add(self, time):
        b = self.buffer
        while len(b) and self.write_ready <= time:
            acq = b.pop(0)
            write_start = max(acq, self.write_ready)
            self.write_ready = write_start + 1040
        overflow = len(b) > 7
        if not overflow:
            b.append(time)
        return not overflow




import logging
from functools import partial
import json
import numpy as np

from qcodes.instrument.channel import InstrumentChannel

from .q1core import Q1Core
from .rt_renderer import Renderer

from qblox_instruments import SequencerState, SequencerStatus, SequencerStatusFlags


class Q1Sequencer(InstrumentChannel):

    # only logged
    _seq_log_only_parameters = [
        # -- only printed:
        'sync_en',
        'nco_phase_offs',
        'marker_ovr_en',
        'marker_ovr_value',
        'cont_mode_en_awg_path0',
        'cont_mode_en_awg_path1',
        'cont_mode_waveform_idx_awg_path0',
        'cont_mode_waveform_idx_awg_path1',
        'upsample_rate_awg_path0',
        'upsample_rate_awg_path1',
        'gain_awg_path0',
        'gain_awg_path1',
        'offset_awg_path0',
        'offset_awg_path1',
        'mixer_corr_phase_offset_degree',
        'mixer_corr_gain_ratio',
        ]
    _seq_log_only_parameters_qrm = [
        'integration_length_acq',
        'phase_rotation_acq',
        'discretization_threshold_acq',
        ]

    def __init__(self, parent, name, sim_type):
        super().__init__(parent, name)
        self._is_qcm = sim_type in ['QCM', 'QCM-RF', 'Viewer']
        self._is_qrm = sim_type in ['QRM', 'QRM-RF', 'Viewer']

        if self._is_qrm:
            log_params = self._seq_log_only_parameters + self._seq_log_only_parameters_qrm
        else:
            log_params = self._seq_log_only_parameters

        for par_name in log_params:
            self.add_parameter(par_name,
                               set_cmd=partial(self._log_set, par_name))

        self.add_parameter('sequence', set_cmd=self.upload)
        self.add_parameter('mod_en_awg', set_cmd=self._set_mod_en_awg)
        self.add_parameter('nco_freq', set_cmd=self._set_nco_freq)
        self.add_parameter('channel_map_path0_out0_en',
                           set_cmd=partial(self._set_channel_map_path_en, 0, 0))
        self.add_parameter('channel_map_path1_out1_en',
                           set_cmd=partial(self._set_channel_map_path_en, 1, 1))
        if self._is_qcm:
            self.add_parameter('channel_map_path0_out2_en',
                               set_cmd=partial(self._set_channel_map_path_en, 0, 2))
            self.add_parameter('channel_map_path1_out3_en',
                               set_cmd=partial(self._set_channel_map_path_en, 1, 3))
        if self._is_qrm:
            self.add_parameter('demod_en_acq', set_cmd=self._set_demod_en_acq)

        self.reset()

    def config(self, name, value):
        if name == 'name':
            self.name = value
            self.rt_renderer.name = value
        elif name == 'max_render_time':
            self.rt_renderer.max_render_time = value
        elif name == 'max_core_cycles':
            self.q1core.max_core_cycles = value

    def reset(self):
        self._mod_en_awg = False
        self._nco_freq = 0.0
        self._demod_en_acq = False

        self.waveforms = {}
        self.weights = {}
        self.acquisition_bins = {}

        self.run_state = 'IDLE'
        self.rt_renderer = Renderer(self.name)
        self.q1core = Q1Core(self.name, self.rt_renderer, self._is_qrm)

    def _log_set(self, name, value):
        logging.info(f'{self.name}: {name}={value}')

    def _set_mod_en_awg(self, value):
        logging.debug(f'{self.name}: mod_en_awg={value}')
        self._mod_en_awg = value

    def _set_nco_freq(self, value):
        logging.info(f'{self.name}: nco_freq={value}')
        self._nco_freq = value

    def _set_demod_en_acq(self, value):
        logging.debug(f'{self.name}: demod_en_acq={value}')
        self._demod_en_acq = value

    def _set_channel_map_path_en(self, path, out, value):
        logging.debug(f'{self.name}: channel_map_path{path}_out{out}_en={value}')
        self.rt_renderer.path_enable(path, out, value)

    def upload(self, file_name):
        with open(file_name) as fp:
            pdict = json.load(fp)
        waveforms = pdict['waveforms']
        weights = pdict['weights']
        acquisitions = pdict['acquisitions']
        program = pdict['program']
        self._set_waveforms(waveforms)
        self._set_weights(weights)
        self._set_acquisition_bins(acquisitions)
        self.q1core.load(program)

    def _set_waveforms(self, waveforms):
        self.waveforms = waveforms
        wavedict = {}
        for name, datadict in waveforms.items():
            index = int(datadict['index'])
            data = np.array(datadict['data'])
            wavedict[index] = data
        self.rt_renderer.set_waveforms(wavedict)

    def _set_weights(self, weights):
        self.weights = weights
        weightsdict = {}
        for name, datadict in weights.items():
            index = int(datadict['index'])
            data = np.array(datadict['data'])
            weightsdict[index] = data
        self.rt_renderer.set_weights(weightsdict)

    def _set_acquisition_bins(self, acq_bins):
        self.acquisition_bins = acq_bins
        bins_dict = {}
        for name, datadict in acq_bins.items():
            index = int(datadict['index'])
            num_bins = int(datadict['num_bins'])
            bins_dict[index] = num_bins
        self.rt_renderer.set_acquisition_bins(bins_dict)

    def get_state(self):
        flags = list(self.q1core.errors | self.rt_renderer.errors)
        return SequencerState(
            SequencerStatus[self.run_state],
            [SequencerStatusFlags[flag.replace(' ','_')] for flag in flags],
        )

    def get_acquisition_state(self):
        if not self._is_qrm:
            raise NotImplementedError('Instrument type is not QRM')
        return True

    def arm(self):
        self.run_state = 'ARMED'

    def run(self):
        self.run_state = 'RUNNING'
        self.rt_renderer.reset()
        self.rt_renderer.set_nco(self._nco_freq, self._mod_en_awg)
        self.q1core.run()
        self.run_state = 'STOPPED'

    def get_acquisition_data(self):
        if not self._is_qrm:
            raise NotImplementedError('Instrument type is not QRM')
        cnt,data = self.rt_renderer.get_acquisition_data()
        result = {}
        for name, datadict in self.acquisition_bins.items():
            index = int(datadict['index'])
            num_bins = int(datadict['num_bins'])
            acq_count = cnt[index]
            path_data = [d/c if c > 0 else float('nan')
                         for d,c in zip(data[index], cnt[index])]

            result[name] = {
                'index':index,
                'acquisition':{
                    'bins':{
                        'integration': {
                            'path0':path_data,
                            'path1':path_data,
                            },
                        'threshold':[0.0]*num_bins,
                        'avg_cnt':acq_count,
                    }
                }}
        return result

    def plot(self):
        self.rt_renderer.plot()

    def print_registers(self, reg_nrs=None):
        self.q1core.print_registers(reg_nrs)
